#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "types.h"

/* defines section        */
#define IOCTL_READ_PORT_UCHAR	 -1673519100 /* CTL_CODE(40000, 0x801, METHOD_BUFFERED, FILE_ANY_ACCESS) */
#define IOCTL_WRITE_PORT_UCHAR	 -1673519096 /* CTL_CODE(40000, 0x802, METHOD_BUFFERED, FILE_ANY_ACCESS) */
#define IDR_DRIVER               110
/* parallel port to use   */
#define PARALLEL_PORT_ADDR       0x378
#define PARALLEL_PORT_NUM        1
/* spi port pins          */
#define SPI_POWER                1
#define SPI_SI                   2
#define SPI_CS                   4
#define SPI_CLK                  8
/* serial flash op-codes  */
#define PAGE_WRITE               0x02
#define STATUS_READ              0x05
#define WRITE_ENABLE             0x06
#define READ_FAST                0x0B
#define BULK_ERASE               0xC7

#define ARGS_REQUIRED            2
#define USAGE                    "serial_flash program.bin"
#define FILE1                    1
#define PAGE_SIZE                256
#define SERIAL_IDLE              0x00
#define MAX_MEMORY_SIZE          2048
/* #define DEBUG                    1 */
#define VERSION                  3    /* V1: 512KB, V2: 1MB flashing V3: fpga flash */  

/* global variables */

/* None */

/* local variables */
static U16  baseport = 0x378;
static S32  sysver = 0;
static HANDLE hdriver = INVALID_HANDLE_VALUE;
static char path[MAX_PATH];

/* prototypes section */
static BOOL OpenPort(int port, int addr);
static S32 SystemVersion(void);
static void InitPort(void);
static U8 prECP(void);
static void pwECP(U8);
static void pwControl(U8 data);
static U8 prStatus(void);
static S32 Opendriver(void);
static U8 Inp32 (U16 addr);
static void Out32 (U16 addr, U8 data);
static int start (void);
static int inst(void);
static U8 mc_read_status(void);
static void end(void);
static void begin(void);
static void swrite(const U8 *data, S32 length);
static U32 push(U32 val, U32 bits);
static void sread(U8 *data, S32 length);
static void mc_program(U32 offset, U8 *buffer, S32 len);
static void mc_read(U32 offset, U8 *buffer, U32 len);
static void mc_erase_bulk(void);

/*************************************************************************/
/*                                                                       */
/* Name:         main                                                    */
/*                                                                       */
/* Description:                                                          */
/*               little program to write data to a serial flash chip     */ 
/*               via parallel port                                       */                  
/*                                                                       */
/*************************************************************************/
int main(int argc, char *argv[])
{
  S32 status = 0;
  S32 size = 0;
  FILE *file_ptr = NULL;
  U8 *program_code = NULL;
  U8 buffer[PAGE_SIZE];
  S32 num_pages = 0;
  S32 page_num = 0;
  S32 a;
  S32 i;
  
  /* check we have a program to load into the serial flash chip */
  if(argc < ARGS_REQUIRED)
  {
      printf("Error wrong number of arguments passed \n");
      printf(USAGE);
      return 0;            
  }
  
  file_ptr = fopen(argv[FILE1], "rb");
  
  if(file_ptr == NULL)
  {
      printf("Cant open %s \n",argv[FILE1]);
      return 0; 
  }
  
  /* get the size of the program */
  fseek(file_ptr, 0, SEEK_END);
  size = ftell(file_ptr) + (PAGE_SIZE - 1);
  fseek(file_ptr, 0, SEEK_SET);
  num_pages = size / PAGE_SIZE;
  
  /* check program can fit inside the serail flash chip */
  if(num_pages >= MAX_MEMORY_SIZE)
  {
      printf("\n Error program greater than 512kb program don't fit\n");
      return 0;                 
  }
    
  /* get some memory to hold the program code */
  program_code = malloc(size);
  if(program_code == NULL)
  {
      fclose(file_ptr);
      printf("Cant malloc memory for program ! \n");
      return 0;             
  }
  
  /* store the program code in memory and close file */
  memset(program_code, 0xFF, size);
  fread(program_code, (size - (PAGE_SIZE - 1)), 1, file_ptr);
  fclose(file_ptr);   
    
  /* try and open a parallel port */  
  if (!OpenPort(PARALLEL_PORT_NUM, PARALLEL_PORT_ADDR))
  {
      free(program_code);                             
      printf("Game over cant get a handle for the parallel port ! \n");
      return TRUE;
  }  
  
  /* change the mode to spp */   
  InitPort();
  
  /* serial flash chip select to idle state (high) */
  Out32(baseport, SPI_CS | SPI_POWER);
  sleep(100);
  
  /* check chip is in idle state before executing instructions */
  status = mc_read_status();
  
  if(status != SERIAL_IDLE)
  {
      printf("serial flash not idle! reporting state 0x%02x\n\n",status);
      free(program_code);
      return TRUE;              
  } 
  else
  {
      printf("serial flash in idle state 0x%02x\n\n",status);
  }  
  
  printf("Bulk erasing flash checking status... \n");
  mc_erase_bulk();
  
  sleep(100);
  
  while (1)
  {
      status = mc_read_status();
      printf("\rstatus = %02x ", status);
	
      /* check to see if serial flash has finshed erasing page */	
      if(0x00 == status)
      {
          break;
      }
      
      sleep(100);
  }
  printf("\n done! \n");
  
  /* program code */
  printf("programming code..\n");
  
  printf("%d pages ...\n", num_pages);
  
  /* start programming from page 0 */
  page_num = 0;
  
  for (i = 0; i < num_pages; i++)
  {
	
      printf("\rprogramming at %4d", (page_num / PAGE_SIZE));
      mc_program(page_num, program_code + (i * PAGE_SIZE), PAGE_SIZE);
		    
      while (1)
	  {
          /* check to see if serial flash has finshed programming page */   
          if(mc_read_status() == 0x00)
          {
              break;
          }
      }  
				
      page_num += PAGE_SIZE ;
  }

  sleep(100);
  
  printf("\ndone, verifying...\n");
  memset(buffer, 0, PAGE_SIZE);
    
  /* check each page make sure everything is lovely */
  for (i = 0; i < num_pages; i++)
  {      
		mc_read((i * PAGE_SIZE), buffer, PAGE_SIZE);
		if(memcmp(buffer, program_code + (i * PAGE_SIZE), PAGE_SIZE) != 0)
		{
			printf("!!! page %d failed to verify \n", i);

            /* find bytes in page that failed to program */
			for(a = 0; a < PAGE_SIZE; a++)
			{
				if(buffer[a] != program_code[(i * PAGE_SIZE) + a])
				{
					printf("%03x: %02x != %02x\n", a, buffer[a], program_code[(i * PAGE_SIZE) + a]);
                }
			}
			
			break;
		}
        else
		{
			printf("\rpage %d ok", i + 1);
		}
  }
  printf("\ndone!");
	
  /* serial flash chip select to idle state (high) */
  Out32(baseport, SPI_CS | SPI_POWER);
  
  sleep(100);
  
  /* release memory before exit */
  free(program_code);
}

/*************************************************************************/
/*                                                                       */
/* Name:         push                                                    */
/*                                                                       */
/* Description:                                                          */
/*               write (length) bytes of data to the parallel port data  */
/*               reg                                                     */                  
/*                                                                       */
/*************************************************************************/
static U32 push(U32 val, U32 bits)
{
	U32 result = 0;
	val <<= (32-bits);
	
	while(bits--)
	{
		int v = SPI_POWER;
		
		/* our we going to transmit a logic one */
		if (val & 0x80000000)
		{
            /* if so set the line high */    
			v |= SPI_SI;
        }
        
        /* shift date to the next bit to send */
		val <<= 1;
		/* shift recieve data ready for next bit */
		result <<= 1;
		
		/* output this bit of data */
        Out32(PARALLEL_PORT_ADDR, v);
		
		/* get the ack lines (MISO) values from status reg */
        if (Inp32(PARALLEL_PORT_ADDR + 1) & 0x40)
        {
			result |= 1;
        }
		
		/* clock the clk line (positive edge) */
        Out32(PARALLEL_PORT_ADDR, v | SPI_CLK);
		Out32(PARALLEL_PORT_ADDR, v);
	}
	
	/* return data from serial flash */
	return result;
}

/*************************************************************************/
/*                                                                       */
/* Name:         swrite                                                  */
/*                                                                       */
/* Description:                                                          */
/*               write (length) bytes of data to the parallel port data  */
/*               reg                                                     */                  
/*                                                                       */
/*************************************************************************/
static void swrite(const U8 *data, S32 length)
{
	while(length--)
	{
		push(*data++, 8);
    }
}

/*************************************************************************/
/*                                                                       */
/* Name:         sread                                                   */
/*                                                                       */
/* Description:                                                          */
/*               read (length) bytes of data to the parallel port status */
/*               reg                                                     */                  
/*                                                                       */
/*************************************************************************/
static void sread(U8 *data, S32 length)
{
	while(length--)
	{
		*data++ = push(0, 8);
    }
}

/*************************************************************************/
/*                                                                       */
/* Name:         begin                                                   */
/*                                                                       */
/* Description:                                                          */
/*               set the chip select line low (active)                   */                  
/*                                                                       */
/*************************************************************************/
static void begin(void)
{
	Out32(PARALLEL_PORT_ADDR, SPI_POWER);
}

/*************************************************************************/
/*                                                                       */
/* Name:         end                                                     */
/*                                                                       */
/* Description:                                                          */
/*               set the chip select line high (idle)                    */                  
/*                                                                       */
/*************************************************************************/
static void end(void)
{
	Out32(PARALLEL_PORT_ADDR, SPI_CS | SPI_POWER);
}

/*************************************************************************/
/*                                                                       */
/* Name:         mc_read_status                                          */
/*                                                                       */
/* Description:                                                          */
/*               read the status register of the serial flash device     */                  
/*                                                                       */
/*************************************************************************/
static U8 mc_read_status(void)
{
	U8 cmd = STATUS_READ;
	begin();
	swrite(&cmd, 1);
	sread(&cmd, 1);
	end();
	return cmd;
}

/*************************************************************************/
/*                                                                       */
/* Name:         mc_erase_bulk                                           */
/*                                                                       */
/* Description:                                                          */
/*               erase the serial flash device to all 0xff               */                  
/*                                                                       */
/*************************************************************************/
static void mc_erase_bulk(void)
{
    U8 cmd; 
    
    /* send write enable */
    cmd = WRITE_ENABLE; 
    begin();
    swrite(&cmd, 1);
    end();
    
    sleep(100);
    
    /* send block erase */
    cmd = BULK_ERASE;
    begin();
    swrite(&cmd, 1);
    end();          
}

/*************************************************************************/
/*                                                                       */
/* Name:         mc_program                                              */
/*                                                                       */
/* Description:                                                          */
/*               write page in serial flash device                       */                  
/*                                                                       */
/*************************************************************************/
static void mc_program(U32 offset, U8 *buffer, S32 len)
{
    U8 cmd[4];   
       
    /* send write enable */
    cmd[0] = WRITE_ENABLE; 
    begin();
    swrite(cmd, 1);
    end();
    
    sleep(100);    
    
    /* send write page */
    cmd[0] = PAGE_WRITE;
    cmd[1] = offset >> 16;
    cmd[2] = offset >> 8;
    cmd[3] = offset;   
    #if DEBUG
	    printf("write page %08x %02x %02x %02x \n", offset, cmd[1], cmd[2], cmd[3]);
	#endif 	
	begin();
	swrite(cmd, 4);
	swrite(buffer, len);
	end();
}

/*************************************************************************/
/*                                                                       */
/* Name:         mc_read                                                 */
/*                                                                       */
/* Description:                                                          */
/*               read page in serial flash device                        */                  
/*                                                                       */
/*************************************************************************/
static void mc_read(U32 offset, U8 *buffer, U32 len)
{
	char cmd[5]={READ_FAST, offset >> 16, offset >> 8, offset, 0};
    #if DEBUG
	    printf("read page %08x %02x %02x %02x \n", offset, cmd[1], cmd[2], cmd[3]);
	#endif 	
	begin();
	swrite(cmd, 5);
	sread(buffer, len);
	end();
}

/*************************************************************************/
/*                                                                       */
/* Name:         OpenPort                                                */
/*                                                                       */
/* Description:                                                          */
/*               0pen a parallel port comm channel                       */                  
/*                                                                       */
/*************************************************************************/
static BOOL OpenPort(int port, int addr)
{
	char *pPortName;
	HANDLE pPortFileHandle = INVALID_HANDLE_VALUE;
	
    switch (port)
	{
	    case 1:	
            pPortName = "LPT1";
		break;
	
        case 2:	
            pPortName = "LPT2";
		break;
		
	    case 3:	
            pPortName = "LPT3";
		break;
		
	    case 4:	
            pPortName = NULL;
		break;
		
	    default:
            return FALSE;	
        break;
	}
	
	printf("Opening parallel port %s \n",pPortName);
	
	if(pPortName != NULL)
	{
		pPortFileHandle = CreateFile(pPortName,GENERIC_READ | GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
		if (pPortFileHandle == INVALID_HANDLE_VALUE)
		{
			baseport = 0;
			printf("Parallel port is in use! \n");
			return FALSE;
		}
	}
	
	baseport = addr;
	sysver = SystemVersion();
	
	printf("On windows system %d \n",sysver);
	
	if (sysver == 1)
	{
		return TRUE;
    }
	else if ((sysver == 2) && (!Opendriver()))
	{
		return TRUE;
    }
	else
	{
		if (pPortFileHandle != INVALID_HANDLE_VALUE)
		{
			CloseHandle(pPortFileHandle);
        }
        
		pPortFileHandle = INVALID_HANDLE_VALUE;
		baseport = 0;
		printf("Failed to load parallel port driver! \n");
		return FALSE;
	}
}

/*************************************************************************/
/*                                                                       */
/* Name:         SystemVersion                                           */
/*                                                                       */
/* Description:                                                          */
/*               Gets windows version info                               */                  
/*                                                                       */
/*************************************************************************/
static S32 SystemVersion(void)
{
	OSVERSIONINFOEX osvi;
	BOOL bOsVersionInfoEx;
	ZeroMemory(&osvi,sizeof(OSVERSIONINFOEX));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	
	if (!(bOsVersionInfoEx = GetVersionEx((OSVERSIONINFO *)&osvi)))
	{
		osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
		if (!GetVersionEx((OSVERSIONINFO *)&osvi))
		{
			  return 0;
        }
	}
	
	switch (osvi.dwPlatformId)
	{      
	    case VER_PLATFORM_WIN32_NT:	
             return 2;
        break;
	    
        case VER_PLATFORM_WIN32_WINDOWS:
             return 1;
        break;
	}
       
	return 0;
}

static S32 Opendriver(void)
{
	hdriver = CreateFile("\\\\.\\hwinterface", GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hdriver != INVALID_HANDLE_VALUE)
	{
		return 0;		/* successfully opened */
    }

	if (start())	/* attempt to start */
	{		/* failed... */
		if (inst())	/* attempt to install */
		{		/* failed */
			printf("Failed to install parallel port I/O driver!");
			return 1;
		}
		if (start())
		{
			printf("Failed to start parallel port I/O driver!");
			return 1;
		}
	}
	
    hdriver = CreateFile("\\\\.\\hwinterface", GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	
    if (hdriver != INVALID_HANDLE_VALUE)
	{
		return 0;
    }
    
	return 1;
}

static int inst(void)
{
	SC_HANDLE Mgr, Ser;
	HRSRC hResource;
	HINSTANCE hInst;

	GetSystemDirectory(path,sizeof(path));
	hResource = FindResource(hInst,MAKEINTRESOURCE(IDR_DRIVER),"bin");
	
	if (hResource)
	{
		HGLOBAL binGlob = LoadResource(hInst,hResource);
		if(binGlob)
		{
			void *binData = LockResource(binGlob);
			if (binData)
			{
				HANDLE file;
				strcat(path,"\\Drivers\\hwinterface.sys");
				file = CreateFile(path,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);
				if(file)
				{
					DWORD size, written;
					size = SizeofResource(hInst,hResource);
					WriteFile(file, binData, size, &written, NULL);
					CloseHandle(file);
				}
			}
		}
	}
	
	Mgr = OpenSCManager(NULL,NULL,SC_MANAGER_ALL_ACCESS);
	
    if (Mgr == NULL)
	{				/* No permission to create service */
		if (GetLastError() == ERROR_ACCESS_DENIED)
        { 
			return 5;	/* error access denied */
        }
	}
	else	Ser = CreateService(Mgr,"hwinterface","hwinterface",SERVICE_ALL_ACCESS,SERVICE_KERNEL_DRIVER,SERVICE_SYSTEM_START,SERVICE_ERROR_NORMAL,"System32\\Drivers\\hwinterface.sys",NULL,NULL,NULL,NULL,NULL);
	CloseServiceHandle(Mgr);
	CloseServiceHandle(Ser);
	return 0;
}

static int start (void)
{
	SC_HANDLE Mgr, Ser;
	Mgr = OpenSCManager(NULL,NULL,SC_MANAGER_ALL_ACCESS);
	
	if (Mgr == NULL)
	{	/* No permission to create service */
		if (GetLastError() == ERROR_ACCESS_DENIED) 
		{
			Mgr = OpenSCManager(NULL,NULL,GENERIC_READ);
			Ser = OpenService(Mgr,"hwinterface",GENERIC_EXECUTE);
			if (Ser)
			{	/* we have permission to start the service */
				if (!StartService(Ser,0,NULL))
				{
					CloseServiceHandle(Ser);
					return 4;	/* we could open the service but unable to start */
				}
			}
		}
	}
	else
	{	/* Successfuly opened Service Manager with full access */
		Ser = OpenService(Mgr,"hwinterface",GENERIC_EXECUTE);
		
		if (Ser)
		{
			if (!StartService(Ser,0,NULL))
			{
				CloseServiceHandle(Ser);
				return 3;	/* opened the Service handle with full access permission, but unable to start */
			}
			else
			{
				CloseServiceHandle(Ser);
				return 0;	/* started */
			}
		}
	}
	
	return 1;
}

/*************************************************************************/
/*                                                                       */
/* Name:         InitPort                                                */
/*                                                                       */
/* Description:                                                          */
/*              try and put the port into spp mode with bidirectional    */                  
/*                                                                       */
/*************************************************************************/
static void InitPort(void)
{
    U8 output;
     
    /* is the parallel ports fifo complete empty */
	if ((prECP() & 0x03) == 0x01)
	{
        /* put port into SPP mode with bidirectional */         
		pwECP(0x34);
		output = prECP();
		if (prECP() == 0x35)
        {
			pwECP((prECP() & 0x1F) | 0x20);	/* try to setting ECP port to bidirectional mode */
            printf("setting ECP port to SPP bidirectional mode  \n");
        }
    }
    
    /* configure the data lines as non-bidirectional, dis-enable IRQ */
	pwControl(0x00);
	/* read status info */
	prStatus();
}

/*************************************************************************/
/*                                                                       */
/* Name:         prECP                                                   */
/*                                                                       */
/* Description:                                                          */
/*               read data from the extended capabilities port           */
/*               (ECP) control register                                  */                  
/*                                                                       */
/*************************************************************************/
static U8 prECP(void)
{
    /* if port didnt open correctly return */     
	if (baseport == 0)
	{
		return 0;
    }
    
    /* read data from the extended control register */
	return Inp32((unsigned short)(baseport+0x402));
}

/*************************************************************************/
/*                                                                       */
/* Name:         pwECP                                                   */
/*                                                                       */
/* Description:                                                          */
/*               write data to the extended capabilities port            */
/*               (ECP) control register                                  */                  
/*                                                                       */
/*************************************************************************/
static void pwECP(U8 data)
{
    /* if port didnt open correctly return */  
	if(baseport == 0)
	{
		return;
    }
    
    /* write data to the extended control register */
	Out32((unsigned short)(baseport+0x402),data);
}

/*************************************************************************/
/*                                                                       */
/* Name:         pwControl                                               */
/*                                                                       */
/* Description:                                                          */
/*               write data to the control port                          */              
/*                                                                       */
/*************************************************************************/
static void pwControl (U8 data)
{
    /* if port didnt open correctly return */   
	if(baseport == 0)
	{
		return;
    }
    
    /* write data to the control register */
	Out32((U16)(baseport+2),data);
}

/*************************************************************************/
/*                                                                       */
/* Name:         prStatus                                                */
/*                                                                       */
/* Description:                                                          */
/*               read ports status register                              */              
/*                                                                       */
/*************************************************************************/
static U8 prStatus(void)
{
    /* if port didnt open correctly return */   
	if(baseport == 0)
	{
		return 0;
    }
    
    /* read the parallel ports status register */
	return Inp32((unsigned short)(baseport+1));
}

/*************************************************************************/
/*                                                                       */
/* Name:         Inp32                                                   */
/*                                                                       */
/* Description:                                                          */
/*               read 8 bits from a ports register                       */              
/*                                                                       */
/*************************************************************************/
static U8 Inp32 (U16 addr)
{
	unsigned int error;
	DWORD BytesReturned;
	unsigned char Buffer[3];

	switch (sysver)
	{
	    case 1:	
            return _inp(addr);	
        break;
        
	    case 2:	
            Buffer[0] = LOBYTE(addr);
		    Buffer[1] = HIBYTE(addr);
		    Buffer[2] = 0;
		    error = DeviceIoControl(hdriver,IOCTL_READ_PORT_UCHAR,&Buffer,2,&Buffer,1,&BytesReturned,NULL);
		    return Buffer[0];	
        break;
	}
	
	return 0;
}

/*************************************************************************/
/*                                                                       */
/* Name:         Out32                                                   */
/*                                                                       */
/* Description:                                                          */
/*               write 8 bits to a ports register                        */              
/*                                                                       */
/*************************************************************************/
static void Out32 (U16 addr, U8 data)
{
	unsigned int error;
	DWORD BytesReturned;        
	BYTE Buffer[3];
	
	switch (sysver)
	{
	    case 1:	
            _outp(addr,data);
        break;
        
	    case 2:	
            Buffer[0] = LOBYTE(addr);
		    Buffer[1] = HIBYTE(addr);
		    Buffer[2] = data;
		    error = DeviceIoControl(hdriver,IOCTL_WRITE_PORT_UCHAR,&Buffer,3,NULL,0,&BytesReturned,NULL);
	    break;
	}
}
